/*
 * EPG functions.
 */
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <time.h>
#include "config.h"
#include "timezone.h"
#include "exedsp.h"
#include "unicode.h"
#include "epg.h"

static int EPG_initialized = 0;

/*
 * Channel and program are constant instances initialized in
 * EPG_init() and so should be static.
 */
static unsigned channel[2];
static unsigned program[2];

#define PROGRAMBUFSIZE 128
#define NAMEBUFSIZE	256

void EPG_init(void)
{
	EPG_initialized = 1;

	_ZN9TCChannelC1Ev(channel);
	_ZN9TCProgramC1Ev(program);
	LOG("EPG initialized, channel[0x%X, 0x%X] program[0x%X, 0x%X]",
		channel[0], channel[1], program[0], program[1]);
}

static int EPG_get_current_time(time_t *unixtime)
{
	int ret = 0;

 	ret = _ZN4TCTv17GetSystemInfoTimeEPmi(g_AppTV, unixtime, 0);

	if (ret == 0)
		LOG("Cannot get the current time");
	else
		LOG("Got unixtime[%d]", (int)(*unixtime));
	return (ret);
}

static int EPG_get_structured_current_time(struct EPG_datetime *datetime)
{
	int ret = 0;

 	ret = _ZN10CTimerUtil10GetSITTimeEPN6PCTime6PTTimeEi(datetime, 0);

	if (ret == 0)
		LOG("Cannot get the current time");
	else
		LOG("Got the current time year[%u] month[%u] day[%u] hour[%u] minute[%u]",
			datetime->year, datetime->month, datetime->day,
			datetime->hour, datetime->minute);
	return (ret);
}

static int EPG_get_current_channel(unsigned *channel)
{
	int ret = 0;

 	ret =_ZN8CAppUtil12GetCurrentChEP9TCChanneli(channel, 0);

	if (ret == 0)
		LOG("Cannot get the current channel");
	else
		LOG("Got the current channel[0x%X, 0x%X]", channel[0], channel[1]);

	return (ret);
}

static int EPG_get_present_program(time_t unixtime, unsigned *channel, unsigned *program)
{
	int ret = 0;

 	ret = _ZN4TCTv17GetPresentProgramEP9TCChanneljP9TCProgram(g_AppTV, channel, &unixtime, program);
	if (ret == 0)
		LOG("Cannot get the current program");
	else
		LOG("Got the current program[0x%X, 0x%X]", program[0], program[1]);

	return (ret);
}

static int EPG_get_program_title(unsigned *program, char *progname, size_t siz)
{
	unsigned short namebuffer[NAMEBUFSIZE];
	unsigned namebuflen = sizeof(namebuffer);
	int ret;

	namebuffer[0] = 0;
	ret = _ZNK9TCProgram8GetTitleEPcPj(program, namebuffer, &namebuflen);
	if (ret == 0) {
		LOG("Cannot get the program title");
		return (0);
	}
	if (namebuflen > sizeof(namebuffer))
		namebuflen = sizeof(namebuffer);
	unicode_convert(namebuffer, namebuflen / sizeof(namebuffer[0]), progname, siz);
	return (ret);
}

void EPG_get_data(char *recpartname, size_t siz)
{
	time_t unixtime = 0;
	char progname[PROGRAMBUFSIZE] = "";
	char timestamp[256] = "";

	*recpartname = '\0';
	if (!EPG_initialized)
		EPG_init();

	if (!EPG_get_current_time(&unixtime) || unixtime == 0) {
		unixtime = time(NULL);
		LOG("Got unixtime[%d]", (int)unixtime);
	}

	if (EPG_get_current_channel(channel) &&
		EPG_get_present_program(unixtime, channel, program) &&
		EPG_get_program_title(program, progname, sizeof(progname))) {

		struct EPG_datetime datetime;

		bzero(&datetime, sizeof(datetime));
		if (EPG_get_structured_current_time(&datetime) && datetime.year >= 2010)
			snprintf(timestamp, sizeof(timestamp), "%4u-%02u-%02u_%02u-%02u",
					 datetime.year, datetime.month, datetime.day,
					 datetime.hour, datetime.minute);
	}
	if (!*timestamp) {
		struct tm *tm;
		int y;

		/*
		 * Try to get the current time by the unixtime.
		 * TODO: fix it. 
		 */
		tm = get_localtime_from_UTC_sytemtime(config.winter_time, config.summer_time);
		y = tm->tm_year;
		if (y < 1900)
			y += 1900;
		if (y >= 2010)
			snprintf(timestamp, sizeof(timestamp), "%4d-%02u-%02u_%02u-%02u",
					 y, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min);
	}

	snprintf(recpartname, siz, "%s%s%s_", timestamp, (*timestamp) ? "_" : "",
			 progname);
}
