/*
 * Graphics functions.
 */
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "config.h"
#include "mod.h"
#include "icons.h"

#define SCREEN_W	(1920/2)
#define SCREEN_H	(1080/2)
#define SCREEN_BPP	32
#define SCREEN_FLAGS	0

#define WARNING_DELAY	2000
#define NO_DELAY	1000
#define DELETED_DELAY	2000

static SDL_Surface *screen = NULL;
static SDL_Surface *icon = NULL;
static SDL_Rect dstrect;

void init_video_mode(void)
{
   if (!config.show_icons) { screen = NULL; return; }
   SDL_Init(SDL_INIT_VIDEO);
   screen = SDL_SetVideoMode(SCREEN_W, SCREEN_H, SCREEN_BPP, SCREEN_FLAGS);
}

int set_icon(const char *pngname)
{
   dstrect.x = config.icon_x;
   dstrect.y = config.icon_y;
   if (screen) {
	   char s[MAX_PATH];

	   snprintf(s, sizeof(s), "%s%s.png", config.path, pngname);
	   icon = IMG_Load(s);
	   if (icon) {
		   Uint32 colorkey = SDL_MapRGB(icon->format, 0, 0, 0);
		   SDL_SetColorKey(icon, SDL_SRCCOLORKEY, colorkey);
		   SDL_BlitSurface(icon, NULL, screen, &dstrect);
		   SDL_Flip(screen);
		   return 1;
	   }
   }
   return 0;
}

void close_video_mode()
{
   if (icon) { SDL_FreeSurface(icon); icon = NULL; }
   SDL_Quit(); screen = NULL;
}

int show_REC(void)
{
   init_video_mode();
   if (screen)
   {
      if (!set_icon(ICON_REC))
      {
         close_video_mode();
         return KEY_DISP_VCR;
      }
      return KEY_NOTHING;
   }
   else close_video_mode();
   return KEY_DISP_VCR;
}
int hide_REC(int recording_deleted)
{
   if (screen)
   {
      close_video_mode();
      if (recording_deleted)
      {
	 init_video_mode();
	 if (screen)
      	 {
            if (set_icon(ICON_DELETED)) SDL_Delay(DELETED_DELAY);
            close_video_mode();
         }
      }
      return KEY_NOTHING;
   }
   else
   return KEY_DISP_TV;
}

int show_FORBIDDEN(void)
{
   if (screen)
   {
      close_video_mode();

      init_video_mode();
      if (screen)
      {
         if (set_icon(ICON_NO)) SDL_Delay(NO_DELAY);
         close_video_mode();
         init_video_mode();
         if (screen)
         {
            set_icon(ICON_REC);
         }
      }
      return KEY_NOTHING;
   }
   else close_video_mode();
   return KEY_DISP_CABLE;
}

int show_WARNING(const char *msg)
{
   init_video_mode();
   if (screen)
   {
      if (set_icon(msg)) SDL_Delay(WARNING_DELAY);
      close_video_mode();
      return KEY_NOTHING;
   }
   else close_video_mode();
   return KEY_DISP_CABLE;
}

int
icon_displayed(void)
{
	return (screen != NULL);
}
