#include <dlfcn.h>
#include <sys/mman.h>
#include "common.h"
#include "config.h"
#include "exedsp.h"
#include "mod.h"
#include "mod_pvr.h"
#include "mod_aspect_ratio.h"
#include "mod_esaving.h"
#include "mod_default_source.h"
#include "mod_audio_switch.h"
#include "mod_keycode.h"

#define KFILL(name) #name, KEY_ ##name
#define KNUMFILL(name) #name, KEY_NUM ##name
#define KSRCFILL(name) #name, KEY_SOURCE_ ##name
struct keyinfo keyinfos[] = {
	{KFILL(RECORD)},
	{KFILL(STOP)},
	{KFILL(PLAY)},
	{KFILL(SCRBACK)},
	{KFILL(PAUSE)},
	{KFILL(SCRFWD)},
	{"FAV.CH", KEY_FAV_CH},
	{KFILL(SOURCE)},
	{KFILL(POWER)},
	{"PUP", KEY_P_UP},
	{"PDOWN", KEY_P_DOWN},
	{"PRE-CH", KEY_PRE_CH},
	{KNUMFILL(0)},
	{KNUMFILL(1)},
	{KNUMFILL(2)},
	{KNUMFILL(3)},
	{KNUMFILL(4)},
	{KNUMFILL(5)},
	{KNUMFILL(6)},
	{KNUMFILL(7)},
	{KNUMFILL(8)},
	{KNUMFILL(9)},
	{KFILL(RET)},
	{KFILL(OK)},
	{KSRCFILL(TV)},
	{"CH.LIST", KEY_CH_LIST},
	{KFILL(CONTENT)},
	{"MEDIA.P", KEY_MEDIA_P},
	{"TTX/MIX", KEY_TTX_MIX},
	{KFILL(VUP)},
	{KFILL(VDOWN)},
	{KFILL(MUTE)},
	{KFILL(MENU)},
	{KFILL(UP)},
	{KFILL(RIGHT)},
	{KFILL(DOWN)},
	{KFILL(LEFT)},
	{KFILL(RED)},
	{KFILL(GREEN)},
	{KFILL(YELLOW)},
	{KFILL(BLUE)},
	{KFILL(TOOLS)},
	{KFILL(INTERNET)},
	{KFILL(EXIT)},
	{KFILL(INFO)},
	{KFILL(GUIDE)},
	{KFILL(SUBT)},
	{NULL, 0}
};

#define MODINIT(name_, init_func_, keypressed_func_)	\
	.name = name_, .init_func = init_func_,				\
		.keypressed_func = keypressed_func_

static struct module_info {
	char *name;
	int (*init_func)(void);
	int (*keypressed_func)(const int *keys, int keys_num, int *key_out);
	int initialized;
} modules[] = {
	{MODINIT("keycode",        keycode_init,        keycode_keypressed)},
	{MODINIT("esaving",        esaving_init,        esaving_keypressed)},
	{MODINIT("aspect_ratio",   aspect_ratio_init,   aspect_ratio_keypressed)},
	{MODINIT("audio_switch",   asw_init,            asw_keypressed)},
	{MODINIT("pvr",            pvr_init,            pvr_keypressed)},
	{MODINIT("default_source", default_source_init, default_source_keypressed)},
	{MODINIT(NULL, NULL, NULL)}
};

static int button_pressed(int key) __attribute__ ((noinline));
static int
button_pressed(int key)
{
	static int keys[MAX_KEYS_HISTORY];
	static int mod_key_pressed = 0;
	static int nkeys = 0;
	struct module_info *mod;
	int rc = MOD_KEYPRESSED_UNKNOWN;
	int res_key;

	if (config.mod_key) {
		if (!mod_key_pressed) {
			if (key == config.mod_key) {
				/* Modifier key pressed first time */
				mod_key_pressed = 1;
				return (KEY_NOTHING);
			}
		} else {
			if (key != config.mod_key)
				key |= MKEY;
			mod_key_pressed = 0;
		}
	}

	memmove(keys + 1, keys, sizeof(keys) - sizeof(keys[0]));
	keys[0] = key;
	if (nkeys < sizeof(keys) / sizeof(keys[0]))
		nkeys++;
	for (mod = modules; mod->name != NULL; mod++) {
		if (!mod->initialized || mod->keypressed_func == NULL)
			continue;
		res_key = 0;
		rc = mod->keypressed_func(keys, nkeys, &res_key);
		if (rc == MOD_KEYPRESSED_PROCESSED)
			break;
	}
	if (rc == MOD_KEYPRESSED_PROCESSED && res_key != 0)
		return ((res_key & MKEY) ? KEY_NOTHING : res_key);
	return ((key & MKEY) ? KEY_NOTHING : key);
}

static void keypressed_injection(int a, int key, int arg_r2) __attribute__ ((noinline));
static void
keypressed_injection(int a, int key, int arg_r2)
{
	// gcc should put "LDMFD SP!, {R4,LR}" on top, if everything goes as planned
	asm volatile(
		"STMFD   SP!, {R0,R2,R3,R12}\n"
	);

	button_pressed(key);
	
	asm volatile(
		"MOV     R1, R0\n" // remapped keycode
		"MOV     R6, R0\n" // remapped keycode
		"LDMFD   SP!, {R0,R2,R3,R12}\n"
		"LDMFD   SP!, {R4,LR}\n"
		"MOV     R7, R2\n"  //This is original command that we overwrite
		"LDR     R3, [R8]\n"  //This is original command that we overwrite
		"LDR     PC, =(_ZN9KeyCommon17SendKeyPressInputEii + 0x18 )\n"	//return after injection
	);
}

static int
keypressed_injector(void)
{
	void *handle;
	void *page;
	unsigned *KeyCommon_SendKeyPressInput;

	if ((handle = dlopen(NULL, RTLD_LAZY | RTLD_GLOBAL)) == NULL) {
		LOG("Cannot dlopen() ourselves");
		return (-1);
	}
	KeyCommon_SendKeyPressInput =
		dlsym(handle, "_ZN9KeyCommon17SendKeyPressInputEii");

	dlclose(handle);

	if (KeyCommon_SendKeyPressInput == NULL) {
		LOG("Cannot find KeyCommon::SendKeyPressInputE(int, int)");
		return (-1);
	}

 	page = (void *)(((unsigned) KeyCommon_SendKeyPressInput) & ~0x0FFF);
	mprotect(page, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	KeyCommon_SendKeyPressInput[4] = 0xE51FF004;
	KeyCommon_SendKeyPressInput[5] = (unsigned)keypressed_injection;
	mprotect(page, 8192, PROT_READ | PROT_EXEC);

	cacheflush(KeyCommon_SendKeyPressInput+4, 8);
	return (0);
}

int
mod_init(const char *path)
{
	struct module_info *mod;
	int ninitialized = 0;

	load_config(path);

	if (load_config(path) < 0 || keypressed_injector() < 0)
		return (-1);

	if ((config.mod_key & MKEY) || config.mod_key < 0) {
		LOG("Invalid modifier key[0x%X], "
			"will use the default value[0x%X]", config.mod_key,
			KEY_GREEN);
		config.mod_key = KEY_GREEN;
	}
	for (mod = modules; mod->name != NULL; mod++) {
		if (mod->init_func != NULL && mod->init_func() == MOD_INIT_IGNORE) {
			mod->initialized = 0;
			continue;
		}
		mod->initialized = 1;
		ninitialized++;
		LOG("Module[%s] initialized", mod->name);
	}
	if (!ninitialized) {
		LOG("There are no modules initialized");
		return (-1);
	}
	LOG("Initialized with %d modules", ninitialized);

	if (!config.debug) {
		LOG("Logging disabled");
		pvr_log_deinit();
	}

	if (config.debug && config.exedsp_debug) {
		pvr_log_stdout();
		/* With printlevel=0xffffffff the TV will hang: too many logging */
		spI_SetPrintLevel(config.exedsp_severity);
		spI_SetPrintModule(config.exedsp_facility);
	}

	return (0);
}

void cacheflush(void *start, size_t size)
{
   // For details, please see the Linux sources, files
   // arch/arm/kernel/traps.c, arch/arm/kernel/entry-common.S and
   // asm-arm/unistd.h

   // This syscall is supported by all 2.4 and 2.6 Linux kernels.  2.2 linux
   // kernels got first support with version 2.2.18.

   asm("mov r0, %0\n"
       "mov r1, %1\n"
       "mov r2, #0\n"
	// EABI or Thumb syscall: syscall number passed in 'r7' (syscall base
	// number is 0x0).  Note that Thumb and EABI are generally not the
	// same.  It just happens that the simple cacheflush syscall doesn't
	// expose any of differences in calling conventions.
       "mov r7, #0xf0000\n"
       "add r7, #0x02\n"
       "swi #0\n" ::
       "g"(start), "g"((long)start+size) :
       "r0", "r1", "r2", "r7");
}
