/*
 * Energy-saving module
 */
#include "config.h"
#include "exedsp.h"
#include "mod.h"
#include "mod_esaving.h"

static int cur_saving_mode = 4;
static int panel_off = 0;

static void
esaving_mute(int key)
{
	static int mflag = 1;

	_ZN8TCWindow9MuteVideoEb(&g_AppWindow, mflag);
	_ZN8TCWindow9MuteVideoEb(&g_AppWindowPIP, mflag);

	LOG("Video mute %s", mflag ? "enabled" : "disabled");

	mflag = !mflag;
}

static void
esaving_check_mode(int key)
{
	unsigned int varr[2];
	unsigned int *rval=varr;
	int retv;

	memset(varr, 0, sizeof(varr));
	retv = _ZN4TCTv9GetOptionEiPi(g_AppTV, TCTV_OPTION_ESAVING, rval);

	/*
     * return: 1 on success
     * Off: 0, Low: 1, Medium: 2, High: 3, Auto: 4, Off: (0)*** 
     */
	LOG("TCTv::GetOption(0xf): ret=0x%X, value=0x%X", retv, *rval);
}

static int
esaving_set_option(int mode)
{
	int retv;

	retv = _ZN4TCTv9SetOptionEii(g_AppTV, TCTV_OPTION_ESAVING,
								 mode);
	LOG("TCTv::SetOption(OPTION_ESAVING): esaving[0x%X] ret[0x%X]",
		mode, retv);
	return (retv);
}

static void
esaving_change_mode(int key)
{
	esaving_set_option(cur_saving_mode);
	if (++cur_saving_mode > 4)
		cur_saving_mode = 0;
}

static void
esaving_panel_off(int key)
{
	if (!panel_off) {
		int ptevent[6];
		esaving_set_option(5);

		/* Special rocessing for panel_off */
		memset(ptevent, 0, sizeof(ptevent));
		ptevent[3] = 0x432;
		ptevent[4] = 1;
		ptevent[5] = 0;
		_ZN4CApp9SendEventEP7PTEvent15DTV_APPLICATIONb(ptevent + 1, 2, 0);

		panel_off = 1;
	} else {
		/* return to the previous saving mode */
		esaving_set_option(cur_saving_mode);
		panel_off = 0;
	}
}

int
esaving_keypressed(const int *keys, int keys_num, int *key_out)
{
	struct {
		int key;
		void (*func)(int key);
	} table[] = {
		{config.esaving_key_mute,      esaving_mute},
		{config.esaving_key_check,     esaving_check_mode},
		{config.esaving_key_change,    esaving_change_mode},
		{config.esaving_key_panel_off, esaving_panel_off},
		{0, NULL}
	}, *cur;

	for (cur = table; cur->func != NULL; cur++) {
		if (keys[0] != cur->key)
			continue;
		cur->func(keys[0]);
		return (MOD_KEYPRESSED_PROCESSED);
	}
	return (MOD_KEYPRESSED_UNKNOWN);
}

int
esaving_init(void)
{
	if (!config.enable_esaving)
		return (MOD_INIT_IGNORE);
	return (MOD_INIT_DONE);
}
