/*
 * UNICODE conversion functions.
 */

#include <arpa/inet.h>
#include <string.h>
#include "config.h"
#include "unicode.h"

#define REPLACEMENT	'_'		// replacement for "bad" characters

/* Unicode -> ascii conversion table */
const static struct {
	unsigned short unicode;
	char *ascii;
} ascii_table[] = {
// Polish
{0x0105, "a"},{0x0104, "A"},{0x0107, "c"},{0x0106, "C"},{0x0119, "e"},{0x0118, "E"},{0x0142, "l"},{0x0141, "L"},
{0x0144, "n"},{0x0143, "N"},{0x00F3, "o"},{0x00D3, "O"},{0x015B, "s"},{0x015A, "S"},{0x017A, "z"},{0x0179, "Z"},
{0x017C, "z"},{0x017B, "Z"},
// Hungarian (2 variants)
{0x0150, "OE"},{0x0151, "oe"},{0x00D5, "OE"},{0x00F5, "oe"},{0x0170, "UE"},{0x0171, "ue"},{0x00DB, "UE"},{0x00FB, "ue"},
// German + Swedish
{0x00E4, "ae"},{0x00C4, "AE"},{0x00F6, "oe"},{0x00D6, "OE"},{0x00FC, "ue"},{0x00DC, "UE"},{0x00DF, "ss"},
// Swedish + Danish
{0x00E5, "a"},{0x00C5, "A"},
// Danish
{0x00E6, "ae"},{0x00C6, "AE"},{0x00F8, "0"},{0x00D8, "0"},
// Russian (thanks to black-byte @ SamyGO forum)
{0x0410, "A"},{0x0430, "a"},{0x0411, "B"},{0x0431, "b"},{0x0412, "V"},{0x0432, "v"},{0x0413, "G"},{0x0433, "g"},
{0x0414, "D"},{0x0434, "d"},{0x0415, "E"},{0x0435, "e"},{0x0401, "E"},{0x0451, "e"},{0x0416, "ZH"},{0x0436, "zh"},
{0x0417, "Z"},{0x0437, "z"},{0x0418, "I"},{0x0438, "i"},{0x0419, "I"},{0x0439, "i"},{0x041A, "K"},{0x043A, "k"},
{0x041B, "L"},{0x043B, "l"},{0x041C, "M"},{0x043C, "m"},{0x041D, "N"},{0x043D, "n"},{0x041E, "O"},{0x043E, "o"},
{0x041F, "P"},{0x043F, "p"},{0x0420, "R"},{0x0440, "r"},{0x0421, "S"},{0x0441, "s"},{0x0422, "T"},{0x0442, "t"},
{0x0423, "U"},{0x0443, "u"},{0x0424, "F"},{0x0444, "f"},{0x0425, "KH"},{0x0445, "kh"},{0x0426, "TS"},{0x0446, "ts"},
{0x0427, "CH"},{0x0447, "ch"},{0x0428, "SH"},{0x0448, "sh"},{0x0429, "SHCH"},{0x0449, "shch"},{0x042A, ""},{0x044A, ""},
{0x042B, "Y"},{0x044B, "y"},{0x042C, ""},{0x044C, ""},{0x042D, "E"},{0x044D, "e"},{0x042E, "YU"},{0x044E, "yu"},
{0x042F, "YA"},{0x044F, "ya"},
// ??? (ìéÌÉ)
{0x011B, "e"},{0x011A, "E"},{0x0115, "e"},{0x0114, "E"},{0x00E9, "e"},{0x00C9, "E"},
// ??? (áÁ)
{0x00E1, "a"},{0x00C1, "A"},
// ??? (úùÚÙ)
{0x016F, "u"},{0x016E, "U"},{0x00FA, "u"},{0x00DA, "U"},
// ??? (??ïÏòÒ)
{0x0165, "t"},{0x0164, "T"},{0x010F, "d"},{0x010E, "D"},{0x0148, "n"},{0x0147, "N"},
// ??? (??èÈøØ??ýÝíÍ)
{0x009A, "s"},{0x008A, "S"},{0x0161, "s"},{0x0160, "S"},{0x010D, "c"},{0x010C, "C"},{0x0159, "r"},{0x0158, "R"},
{0x009E, "z"},{0x008E, "Z"},{0x017E, "z"},{0x017D, "Z"},{0x00FD, "y"},{0x00DD, "Y"},{0x00ED, "i"},{0x00CD, "I"},
{0x0000, ""} // end of table
};

static unsigned utf8_table1[] = {0x7f, 0x7ff, 0xffff, 0x1fffff, 0x3ffffff, 0x7fffffff};
#define UTF8_TABLE1_SIZE (sizeof(utf8_table1) / sizeof(utf8_table1[0]))

static unsigned char utf8_table2[] = {0, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc};

static int
bad_char_check(char sym)
{
	char *bad = "/\"?:<>*\\|";

	return (strchr(bad, sym) != NULL);
}

static int
unicode2ascii(unsigned short *unicode, int nchars, char *ascii, size_t asciisize)
{
	unsigned short *psrc, *plast;
	unsigned short uchar;
	char *pdst, *pconv;
	int found;

	psrc = unicode;
	plast = unicode + nchars;
	pdst = ascii;
	for (; psrc < plast && (uchar = ntohs(*psrc)) != 0;
		 psrc++, asciisize--) {

		if (asciisize <= 1)
			break;

		found = 0;
		if (config.unicode == CFG_UNICODE_ASCII_TRANSCODE) {
			int i;

			for (i = 0; ascii_table[i].unicode != 0; i++) {
				if (uchar != ascii_table[i].unicode)
					continue;
				pconv = ascii_table[i].ascii;
				while (*pconv)
					*(pdst++) = *(pconv++);
				found = 1;
				break;
			}
		}
		if (!found) {
			if ((uchar & 0xff00) || bad_char_check(uchar))
				*(pdst++) = REPLACEMENT;
			else
				*(pdst++) = uchar;
		}
	}
	*pdst = '\0';

	return 0;
}

static int
unicode2utf8(unsigned short *unicode, int nchars, char *utf8, unsigned utf8size)
{
	unsigned uchar;
	unsigned cchar;
	int i, j;

	for (cchar = 0; cchar < nchars && (uchar = ntohs(unicode[cchar])) != 0; cchar++) {
	    if (utf8size <= UTF8_TABLE1_SIZE)
			break;

	    for (i = 0; i < UTF8_TABLE1_SIZE; i++)
			if (uchar <= utf8_table1[i]) break;
	    utf8 += i;
	    utf8[1] = '\0'; /* for debug logging only */
	    for (j = i; j > 0; j--) {
			*utf8-- = 0x80 | (uchar & 0x3F);
			uchar >>= 6;
	    }
	    *utf8 = utf8_table2[i] | uchar;

		/* Replace 'bad' one-byte characters by REPLACEMENT */
	    if (i == 0 && bad_char_check(*utf8))
			*utf8 = REPLACEMENT;

	    i++;
	    utf8 += i;
	    utf8size -= i;
	}
	*utf8 = '\0';

	return (0);
}

int
unicode_convert(unsigned short *unicode, int nchars, char *buf, unsigned bufsize)
{
	if (config.unicode == CFG_UNICODE_UTF8)
		return (unicode2utf8(unicode, nchars, buf, bufsize));
	else
		return (unicode2ascii(unicode, nchars, buf, bufsize));
}
